# This is for directly manipulating the yaml file, as well as
# storing configuration for floatyhelper itself.
require 'yaml'
require 'etc'

class Config
  DEFAULT_CONF_FILE = <<~EOT
    ---
    config: {}
    vms: {}
    snapshots: {}
  EOT
                      .freeze

  VALID_SETTINGS = {
    'increaselife' => {
      'description' => 'Number of hours beyond the service default to increase the VM lifetime when adding a VM to floatyhelper.',
      'default' => 0,
    },
    'vertical_snapshot_status' => {
      'description' => 'When true, clear the screen and show the snapshot in progress status vertically. Otherwise, show status on a single line.',
      'default' => true,
    },
  }.freeze

  attr_reader :VALID_SETTINGS

  def self.fhfile
    "#{Etc.getpwuid.dir}/.floatyhelper.yaml"
  end

  def self.load_data
    File.write(fhfile, DEFAULT_CONF_FILE) unless File.exist?(fhfile)
    data = YAML.load_file(fhfile)
    # If someone is using an older version, ensure the conf file has
    # all the correct top-level keys. Only write it back if we modify
    # the data to add a top-level key.
    writeback = false
    ['config', 'vms', 'snapshots'].each do |key|
      if !data.keys.include?(key)
        data[key] = {}
        writeback = true
      end
    end

    # Config setting defaults
    VALID_SETTINGS.each do |setting, info|
      writeback = true if data['config'][setting].nil?
      data['config'][setting] ||= info['default']
    end

    write_data(data) if writeback
    data
  end

  # It is up to the calling function to ensure the data object
  # being passed in is a properly formatted hash. It should only be
  # used to modify after reading the current state with load_data.
  def self.write_data(data)
    File.write(fhfile, data.to_yaml)
  end

  # NOTE: Anything currently set with 'floatyhelper config set' will return a string,
  # regardless of its intended type. Need to add typing one of these days.
  def self.get_config_setting(setting)
    data = load_data
    data['config'][setting]
  end

  def self.set_config_setting(setting, value)
    data = load_data
    data['config'][setting] = value
    write_data(data)
  end
end
